var consoleService = Components.classes['@mozilla.org/consoleservice;1'].getService(Components.interfaces.nsIConsoleService);

/*

CHANGELOG
=========

1.0 beta 2
----------

-Bugfix: getCellProperties for focused tab does not keep for closed then reopened window lists
-Bugfix: Focus, Close, moveTabs functions errored on command without selection
-Bugfix: Closing tabs sometimes would kill entire window if other tabs still open
-Bugfix: focusMe would generate "depreciated property" JS error
-Enhancement: Added focus tab on double-click tree row functionality

*/


var debug = true;
var tabListData = new Array();
var visibleTabData = new Array();
var windowCount, treeView;

function displayOpenTabs()
{
	gettabListData();
	populateTreeView();
	setTreeView();
}

function focusMe()
{
	var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator);
	var enumerator = wm.getEnumerator("");

	while(enumerator.hasMoreElements())	{
		var win = enumerator.getNext();
		if (win.location == "chrome://tabmanager/content/tabmanager.xul")
			win.focus();
	}
}

function moveTabLeft()
{
	var selectedTab = document.getElementById('tabList').currentIndex;
	if (selectedTab == -1) return;
	
	var win = visibleTabData[selectedTab][6];
	win.gBrowser.moveTabTo(win.gBrowser.mTabs[visibleTabData[selectedTab][3]], visibleTabData[selectedTab][3]-1);
	displayOpenTabs();
	document.getElementById('tabList').view.selection.select(selectedTab-1);
}

function moveTabRight()
{
	var selectedTab = document.getElementById('tabList').currentIndex;
	if (selectedTab == -1) return;
	
	var win = visibleTabData[selectedTab][6];
	win.gBrowser.moveTabTo(win.gBrowser.mTabs[visibleTabData[selectedTab][3]], visibleTabData[selectedTab][3]+1);
	displayOpenTabs();
	document.getElementById('tabList').view.selection.select(selectedTab+1);
}

function moveTabNextWin()
{
	var selectedTab = document.getElementById('tabList').currentIndex;
	if (selectedTab == -1) return;
	
	var thisWinNum = visibleTabData[selectedTab][2];
	var thisTabNum = visibleTabData[selectedTab][3];

	var thisWin = visibleTabData[selectedTab][6];
	var thisBrowser = thisWin.gBrowser.getBrowserAtIndex(thisTabNum);
	var nextWin = findNextWin(thisWinNum);

	moveTabToWin(thisBrowser, nextWin);
	window.setTimeout(function() {focusMe();}, 0);
	window.setTimeout(function() {displayOpenTabs();}, 300);
	window.setTimeout(function() {document.getElementById('tabList').view.selection.select(selectedTab);}, 500);
}

function moveTabPrevWin()
{
	var selectedTab = document.getElementById('tabList').currentIndex;
	if (selectedTab == -1) return;
	
	var thisWinNum = visibleTabData[selectedTab][2];
	var thisTabNum = visibleTabData[selectedTab][3];

	var thisWin = visibleTabData[selectedTab][6];
	var thisBrowser = thisWin.gBrowser.getBrowserAtIndex(thisTabNum);
	var prevWin = findPrevWin(thisWinNum);

	moveTabToWin(thisBrowser, prevWin);
	window.setTimeout(function() {focusMe();}, 0);
	window.setTimeout(function() {displayOpenTabs();}, 300);
	window.setTimeout(function() {document.getElementById('tabList').view.selection.select(selectedTab+1);}, 500);
}

function moveTabToWin(browser, win)
{
	var clonedHistory = copyHistory(browser.sessionHistory);
	if (browser.contentWindow && browser.contentWindow.document.forms) {
		var clonedForms = copyForms(browser.contentWindow.document.forms);
	}
	else {
		var clonedForms = null;
	}

	var scrollPosX = browser.contentWindow.scrollX;
	var scrollPosY = browser.contentWindow.scrollY;
	//var zoomFactor = ZoomManager.prototype.getInstance().textZoom;

	var newTab = win.gBrowser.addTab();
	win.gBrowser.selectedTab = newTab;

	var openedBrowser = win.gBrowser.getBrowserForTab(newTab);
	cloneHistory(openedBrowser.webNavigation, clonedHistory);

	if (clonedForms) {
		window.setTimeout(function() {cloneForms(openedBrowser.contentWindow.document.forms, clonedForms);}, 0);
	}
	window.setTimeout(function() {openedBrowser.contentWindow.scrollTo(scrollPosX, scrollPosY);}, 0);
}

function focusTab()
{
	var selectedTab = document.getElementById('tabList').currentIndex;
	if (selectedTab == -1) return;

	var win = visibleTabData[selectedTab][6];
	var newTabFocus = visibleTabData[selectedTab][7];

	win.gBrowser.selectedTab = newTabFocus;
	displayOpenTabs();
	document.getElementById('tabList').view.selection.select(selectedTab);
	window.setTimeout(function() {focusMe();}, 0);
}

// Does not work when only one tab in window
function closeTab()
{
	var selectedTab = document.getElementById('tabList').currentIndex;
	if (selectedTab == -1) return;

	var thisWin = visibleTabData[selectedTab][6];
	var thisTab = thisWin.gBrowser.mTabs[visibleTabData[selectedTab][3]];

	// Avoids index out of bounds errors with short circuit logic
	if (selectedTab == visibleTabData.length-1 && visibleTabData[selectedTab][3] == 0) {
		thisWin.close();
	}
	else if(selectedTab == visibleTabData.length-1 && visibleTabData[selectedTab][3] > 0) {
		thisWin.gBrowser.removeTab(thisTab);
	}
	else if (visibleTabData[selectedTab][3] == 0 && visibleTabData[selectedTab+1][2] > visibleTabData[selectedTab][2]) {
		thisWin.close();
	}
	else {
		thisWin.gBrowser.removeTab(thisTab);
	}
	displayOpenTabs();

	if (selectedTab >= visibleTabData.length)
		selectedTab = visibleTabData.length - 1;

	document.getElementById('tabList').view.selection.select(selectedTab);
	window.setTimeout(function() {focusMe();}, 0);
}

function newTab()
{
	var selectedTab = document.getElementById('tabList').currentIndex;
	if (selectedTab == -1) return;
	
	var thisWin = visibleTabData[selectedTab][6];
	var newTab = thisWin.gBrowser.addTab();
	thisWin.gBrowser.selectedTab = newTab;

	window.setTimeout(function() {focusMe();}, 0);
	window.setTimeout(function() {displayOpenTabs();}, 300);
	//window.setTimeout(function() {document.getElementById('tabList').view.selection.select(selectedTab+1);}, 500);
}

function newWindow()
{
	window.open();
	window.setTimeout(function() {focusMe();}, 0);
	window.setTimeout(function() {displayOpenTabs();}, 300);
	//window.setTimeout(function() {document.getElementById('tabList').view.selection.select(selectedTab+1);}, 500);
}

function tabListSelect()
{
	var lastTab = -1;
	var selectedTab = document.getElementById('tabList').currentIndex;
	
	// If selection is a tab (not a window)
	if (visibleTabData[selectedTab][3] != -1)
	{
		var lastTab = findLastTabInWindow(selectedTab);
		
		// Enable/disable left & right buttons
		if (visibleTabData[selectedTab][3] > 0) {
			document.getElementById('left').disabled = false;
		}
		else {
			document.getElementById('left').disabled = true;
		}
		if (selectedTab != lastTab) {
			document.getElementById('right').disabled = false;
		}
		else {
			document.getElementById('right').disabled = true;
		}

		// Enable/disable previous & next window buttons
		if (visibleTabData[selectedTab][2] == 0) {
			document.getElementById('previouswin').disabled = true;
		}
		else {
			document.getElementById('previouswin').disabled = false;
		}
		if (windowCount <= visibleTabData[selectedTab][2]) {
			document.getElementById('nextwin').disabled = true;
		}
		else {
			document.getElementById('nextwin').disabled = false;
		}

		document.getElementById('focustab').disabled = false;
		document.getElementById('closetab').disabled = false;

		// Update "statusbar" caption
		document.getElementById('selectedTab').value = "Window: " + visibleTabData[selectedTab][2] + "  Tab: " + visibleTabData[selectedTab][3];
	}
	// If selection is a window, disable all buttons
	else {
		document.getElementById('left').disabled = true;
		document.getElementById('right').disabled = true;
		document.getElementById('previouswin').disabled = true;
		document.getElementById('nextwin').disabled = true;
		document.getElementById('closetab').disabled = true;
		document.getElementById('focustab').disabled = true;
		document.getElementById('selectedTab').value = "Window: " + visibleTabData[selectedTab][2];
	}
}

function doubleClickList()
{
	tabListSelect();
	focusTab();
}

function gettabListData()
{
	var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator);
	var enumerator = wm.getEnumerator("navigator:browser");
	var thisTabData;

	windowCount = -1;
	tabListData.length = 0;
	visibleTabData.length = 0;

	while(enumerator.hasMoreElements())
	{
		var win = enumerator.getNext();
		var numTabs = win.gBrowser.mPanelContainer.childNodes.length;
		var thisTitle = (win.gBrowser.selectedBrowser.contentTitle) ?
			win.gBrowser.selectedBrowser.contentTitle : "Untitled";
		
		//print(win.gBrowser.selectedBrowser);

		windowCount++;
		thisTabData = new Array(9);
		thisTabData[0] = "Window " + windowCount + " [ " + thisTitle + " ]";
		thisTabData[1] = "";
		thisTabData[2] = windowCount;
		thisTabData[3] = -1;
		thisTabData[4] = true;		// Is container
		thisTabData[5] = true;		// Is container open?
		thisTabData[6] = win;
		thisTabData[7] = null;
		thisTabData[8] = false;

		tabListData.push(thisTabData);
		visibleTabData.push(thisTabData);

		for(var tabNum=0; tabNum<numTabs; tabNum++)
		{
			//var browser = win.gBrowser.mTabs[tabNum];
			var browser = win.gBrowser.getBrowserAtIndex(tabNum);
			var URI = (browser.currentURI.spec) ? browser.currentURI.spec : "(Untitled)";
			var pageTitle = (browser.contentTitle) ? browser.contentTitle : "(Untitled)";

			var thisTabData = new Array(9);
			thisTabData[0] = pageTitle;
			thisTabData[1] = URI;
			thisTabData[2] = windowCount;
			thisTabData[3] = tabNum;
			thisTabData[4] = false;		// Is container
			thisTabData[5] = false;		// Is container open?
			thisTabData[6] = win;
			thisTabData[7] = win.gBrowser.mTabs[tabNum];
			if (browser == win.gBrowser.selectedBrowser)
				thisTabData[8] = true;
			else
				thisTabData[8] = false;

			tabListData.push(thisTabData);
			visibleTabData.push(thisTabData);
		}
	}
}

function populateTreeView()
{
	treeView = {
		get rowCount() { return visibleTabData.length; },
		getCellText : function(row, column) {
			if (column.id == "titleCol")
				return visibleTabData[row][0];
			else if (column.id == "uriCol")
				return visibleTabData[row][1];
			else
				return "error"
		},
		setTree: function(treebox) { this.treebox = treebox; },
		isSeparator: function(row) { return false; },
		isSorted: function() { return false; },
		isEditable: function(idx, column) { return false; },
		getImageSrc: function(row, col) { return null; },
		getProgressMode: function(idx,column) {},
		getCellValue: function(idx, column) {},
		cycleHeader: function(col, elem) {},
		selectionChanged: function() {},
		cycleCell: function(idx, column) {},
		performAction: function(action) {},
		performActionOnCell: function(action, index, column) {},
		getRowProperties: function(row, props) {},
		getCellProperties: function(row, col, props) {
			var aserv = Components.classes['@mozilla.org/atom-service;1'].getService(Components.interfaces.nsIAtomService);

			if (visibleTabData[row][8]) {
				props.AppendElement(aserv.getAtom("currentTab"));
			}
			if (visibleTabData[row][3] == -1) {
				props.AppendElement(aserv.getAtom("windowTitle"));
			}
		},
		getColumnProperties: function(column, prop) {},
		isContainer: function(row) { return visibleTabData[row][4];	},
		isContainerOpen: function(row) { return visibleTabData[row][5]; },
		isContainerEmpty: function(row) { return false; },
		getLevel: function(row) {
			if (this.isContainer(row)) return 0;
			return 1;
		},
		hasNextSibling: function(row, after) {
			var thisLevel = this.getLevel(row);
			for (var t = row+1; t < visibleTabData.length; t++)
			{
				var nextLevel = this.getLevel(t);
				if (nextLevel == thisLevel)	return true;
				else if (nextLevel < thisLevel) return false;
			}
			return false;
		},
		getParentIndex: function(row) {
			if (this.isContainer(row)) return -1;
			for (var t = row-1; t >= 0; t--) {
				if (this.isContainer(t)) return t;
			}
			return -1;
		},
		getLastSibling: function(row) {
			if (this.isContainer(row)) return -1;
			for (var t = row; t < visibleTabData.length; t++) {
				if (this.isContainer(t)) return t-1;
			}
			if (t == visibleTabData.length) return t-1;
			return -1;
		},
		toggleOpenState: function(row) {
			var item = visibleTabData[row];
			if (!item[4]) return;

			// Close tree
			if (item[5]) {
				item[5] = false;
				var thisLevel = this.getLevel(row);
				var deletecount = 0;

				for (var t = row + 1; t < visibleTabData.length; t++) {
					if (this.getLevel(t) > thisLevel) deletecount++;
					else break;
				}
				if (deletecount) {
					visibleTabData.splice(row + 1, deletecount);
					this.treebox.rowCountChanged(row + 1, -deletecount);
				}
			}
			// Open tree
			else {
				item[5] = true;

				windowNum = visibleTabData[row][2];
				for (var tabListStart = 0; tabListStart < tabListData.length; tabListStart++) {
					if (tabListData[tabListStart][2] == windowNum) {
						tabListStart++; break;
					}
				}
				if (tabListStart == tabListData.length) {
					alert("error");	return;
				}

				for (var tabListEnd = tabListStart; tabListEnd < tabListData.length; tabListEnd++) {
					if (tabListData[tabListEnd][2] > windowNum) {
						tabListEnd--; break;
					}
					if (tabListEnd == tabListData.length-1) break;
				}

				for (var i = tabListStart; i <= tabListEnd; i++) {
					var thisTabData = new Array(9);
					for (var j = 0; j < 9; j++) {
						thisTabData[j] = tabListData[i][j];
					}
					visibleTabData.splice(row + i + 1 - tabListStart, 0, thisTabData);
				}
				this.treebox.rowCountChanged(row + 1, tabListEnd-tabListStart+1);
			}
		}
	};
}

function setTreeView()
{
	document.getElementById('tabList').view=treeView;
}